import functools
import logging
from typing import List, Optional, Set, Tuple

import sqlglot
from pydantic import BaseModel
from sqlglot import exp, optimizer
from sqlglot.errors import ParseError

from quollio_core.helper.core import new_global_id
from quollio_core.profilers.lineage import LineageInput, LineageInputs

logger = logging.getLogger(__name__)


@functools.total_ordering
class _FrozenModel(BaseModel, frozen=True):
    def __lt__(self, other: "_FrozenModel") -> bool:
        for field in self.__fields__:
            self_v = getattr(self, field)
            other_v = getattr(other, field)
            if self_v != other_v:
                return self_v < other_v

        return False


class Table(_FrozenModel):
    db: Optional[str] = None
    db_schema: Optional[str] = None
    table: str

    @classmethod
    def from_sqlglot_table(cls, table: exp.Table, db: str = None, schema: str = None) -> "Table":
        database = table.catalog
        db_schema = table.db
        if database == "":
            database = db or database
        if db_schema == "":
            db_schema = schema or db_schema
        return cls(db=database, db_schema=db_schema, table=table.this.name)


class SQLLineage:
    def __init__(self):
        self.dialects_use_uppercase_normally = ["oracle", "snowflake"]

    def get_table_level_lineage_source(
        self,
        sql: str,
        dialect: str,
        src_db: str = None,
        src_schema: str = None,
        dest_db: str = None,
        dest_schema: str = None,
    ) -> Tuple[Set[Table], Table]:
        try:
            statement: sqlglot.Expression = sqlglot.parse_one(sql=sql, error_level=sqlglot.ErrorLevel.RAISE)
        except ParseError as e:
            logger.error("SQL parse error.\nSQL statement:{sql} \nError:{err}\n".format(sql=sql, err=e))
            raise

        if dialect in self.dialects_use_uppercase_normally:
            src_db = src_db.upper() if src_db is not None else None
            src_schema = src_schema.upper() if src_schema is not None else None
            dest_db = dest_db.upper() if dest_db is not None else None
            dest_schema = dest_schema.upper() if dest_schema is not None else None

        # MEMO: Complement sql with dialect, source database and source schema info.
        optimized_stmt: sqlglot.Expression = optimizer.qualify.qualify(
            statement,
            dialect=dialect,
            catalog=src_db,
            db=src_schema,
            qualify_columns=False,
            validate_qualify_columns=False,
            identify=False,
        )

        orig_dest_table = Table(table="")
        dest_table = Table(table="")
        for expr in optimized_stmt.find_all(exp.Create, exp.Insert, exp.Update, exp.Delete, exp.Merge):
            if isinstance(expr.this, exp.Table):
                orig_dest_table = Table.from_sqlglot_table(
                    table=expr.this
                )  # MEMO: use this to remove duplication from source tables.
                dest_table = Table.from_sqlglot_table(table=expr.this, db=dest_db, schema=dest_schema)

        cte_tables: Set[Table] = set()
        for cte in optimized_stmt.find_all(exp.CTE):
            cte_table = Table(db=None, schema=None, table=cte.alias_or_name)
            cte_tables.add(cte_table)

        source_tables: Set[Table] = set()
        for expr in optimized_stmt.find_all(exp.Table):
            src_orig_table = Table.from_sqlglot_table(expr)
            if src_orig_table != orig_dest_table and src_orig_table not in cte_tables:
                src_table = Table.from_sqlglot_table(expr, db=src_db, schema=src_schema)
                source_tables.add(src_table)

        return source_tables, dest_table

    def gen_lineage_input(
        self, tenant_id: str, endpoint: str, src_tables: Set[Table], dest_table: Table
    ) -> List[LineageInputs]:
        lineage_input = LineageInput(upstream=[])
        downstream_global_id_arg = "{db}{schema}{table}".format(
            db=dest_table.db, schema=dest_table.db_schema, table=dest_table.table
        )
        downstream_global_id = new_global_id(
            tenant_id=tenant_id, cluster_id=endpoint, data_id=downstream_global_id_arg, data_type="table"
        )

        for src_table in src_tables:
            upstream_global_id_arg = "{db}{schema}{table}".format(
                db=src_table.db, schema=src_table.db_schema, table=src_table.table
            )
            upstream_table_global_id = new_global_id(
                tenant_id=tenant_id, cluster_id=endpoint, data_id=upstream_global_id_arg, data_type="table"
            )
            lineage_input.upstream.append(upstream_table_global_id)

        lineage_inputs = LineageInputs(
            downstream_global_id=downstream_global_id,
            downstream_database_name=dest_table.db,
            downstream_schema_name=dest_table.db_schema,
            downstream_table_name=dest_table.table,
            downstream_column_name="",
            upstreams=lineage_input,
        )
        return lineage_inputs



import argparse
import json

from google.auth.credentials import Credentials

from quollio_core.helper.env_default import env_default
from quollio_core.helper.log_utils import configure_logging, error_handling_decorator, logger
from quollio_core.profilers.bigquery import bigquery_table_lineage, bigquery_table_stats
from quollio_core.repository import qdc
from quollio_core.repository.bigquery import BigQueryClient, get_credentials, get_org_id


def initialize_credentials(credentials_json: str) -> Credentials:
    return get_credentials(json.loads(credentials_json))


def initialize_org_id(credentials_json: str) -> str:
    return get_org_id(json.loads(credentials_json))


def initialize_bq_client(credentials: Credentials, project_id: str) -> BigQueryClient:
    return BigQueryClient(credentials=credentials, project_id=project_id)


@error_handling_decorator
def load_lineage(
    tenant_id: str,
    project_id: str,
    regions: list,
    org_id: str,
    credentials: Credentials,
    qdc_client: qdc.QDCExternalAPIClient,
) -> None:
    logger.info("Loading lineage data.")
    bigquery_table_lineage(
        qdc_client=qdc_client,
        tenant_id=tenant_id,
        project_id=project_id,
        regions=regions,
        credentials=credentials,
        org_id=org_id,
    )
    logger.info("Lineage data loaded successfully.")


@error_handling_decorator
def load_stats(
    conn: BigQueryClient,
    tenant_id: str,
    org_id: str,
    qdc_client: qdc.QDCExternalAPIClient,
    dataplex_stats_tables: list,
) -> None:
    logger.info("Loading statistics data.")
    bigquery_table_stats(
        bq_client=conn,
        qdc_client=qdc_client,
        tenant_id=tenant_id,
        org_id=org_id,
        dataplex_stats_tables=dataplex_stats_tables,
    )
    logger.info("Statistics data loaded successfully.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="Quollio Intelligence Agent for BigQuery",
        description="Load lineage and stats to Quollio from BigQuery using Dataplex and BigQuery APIs",
        epilog="Copyright (c) 2024 Quollio Technologies, Inc.",
    )
    parser.add_argument(
        "commands",
        choices=["load_lineage", "load_stats"],
        type=str,
        nargs="+",
        help="""
        The command to execute.
        'load_lineage': Load lineage data from created views to Quollio,
        'load_stats': Load stats from created views to Quollio,
        """,
    )
    parser.add_argument(
        "--log_level",
        type=str,
        choices=["debug", "info", "warn", "error", "none"],
        action=env_default("LOG_LEVEL"),
        default="info",
        required=False,
        help="The log level for dbt commands. Default value is info",
    )
    parser.add_argument(
        "--tenant_id",
        type=str,
        action=env_default("TENANT_ID"),
        required=False,
        help="The tenant id (company id) where the lineage and stats are loaded",
    )
    parser.add_argument(
        "--project_id",
        type=str,
        default=None,
        required=False,
        help="Project ID of the BigQuery project to load lineage and stats from (default is loaded from credentials)",
    )
    parser.add_argument(
        "--regions",
        type=str,
        action=env_default("GCP_REGIONS"),
        required=True,
        help="Comma-separated list of regions BigQuery data is in",
    )
    parser.add_argument(
        "--credentials_json",
        type=str,
        action=env_default("GOOGLE_APPLICATION_CREDENTIALS"),
        required=True,
        help="Credentials JSON",
    )
    parser.add_argument(
        "--api_url",
        type=str,
        action=env_default("QDC_API_URL"),
        required=False,
        help="The base URL of Quollio External API",
    )
    parser.add_argument(
        "--client_id",
        type=str,
        action=env_default("QDC_CLIENT_ID"),
        required=False,
        help="The client id that is created on Quollio console to let clients access Quollio External API",
    )
    parser.add_argument(
        "--client_secret",
        type=str,
        action=env_default("QDC_CLIENT_SECRET"),
        required=False,
        help="The client secret that is created on Quollio console to let clients access Quollio External API",
    )

    parser.add_argument(
        "--dataplex_stats_tables",
        type=str,
        action=env_default("DATAPLEX_STATS_TABLES"),
        required=False,
        help="Comma-separated list of dataplex stats tables - <project_id>.<dataset_id>.<table_id>",
    )

    args = parser.parse_args()

    # Validate that dataplex_stats_tables is provided if load_stats is in commands
    if "load_stats" in args.commands and not args.dataplex_stats_tables:
        parser.error("--dataplex_stats_tables is required when 'load_stats' command is used")

    configure_logging(args.log_level)

    credentials = initialize_credentials(args.credentials_json)
    org_id = initialize_org_id(args.credentials_json)
    qdc_client = qdc.initialize_qdc_client(args.api_url, args.client_id, args.client_secret)
    bq_client = initialize_bq_client(credentials, args.project_id)
    if args.project_id is None:
        args.project_id = json.loads(args.credentials_json)["project_id"]
    regions = args.regions.split(",")

    if "load_lineage" in args.commands:
        load_lineage(
            tenant_id=args.tenant_id,
            project_id=args.project_id,
            regions=regions,
            org_id=org_id,
            credentials=credentials,
            qdc_client=qdc_client,
        )

    if "load_stats" in args.commands:
        tables = args.dataplex_stats_tables.split(",")
        load_stats(
            conn=bq_client,
            tenant_id=args.tenant_id,
            org_id=org_id,
            qdc_client=qdc_client,
            dataplex_stats_tables=tables,
        )



import json
from dataclasses import asdict, dataclass
from typing import Dict, List, Tuple, Union

from quollio_core.helper.core import new_global_id


@dataclass
class LineageInput:
    upstream: List[str]

    def as_dict(self) -> Dict[str, str]:
        return asdict(self)


@dataclass
class LineageInputs:
    downstream_global_id: str
    downstream_database_name: str
    downstream_schema_name: str
    downstream_table_name: str
    downstream_column_name: str
    upstreams: LineageInput


def gen_table_lineage_payload(
    tenant_id: str, endpoint: str, tables: List[Dict[str, Union[Dict[str, str], str]]]
) -> List[LineageInputs]:
    payload = list()
    for table in tables:
        downstream_table_fqdn = table["DOWNSTREAM_TABLE_NAME"].split(".")
        if len(downstream_table_fqdn) != 3:
            continue
        else:
            global_id_arg = "{db}{schema}{table}".format(
                db=downstream_table_fqdn[0], schema=downstream_table_fqdn[1], table=downstream_table_fqdn[2]
            )
            downstream_table_global_id = new_global_id(
                tenant_id=tenant_id, cluster_id=endpoint, data_id=global_id_arg, data_type="table"
            )
            lineage_input = LineageInput(upstream=[])
            for upstream_table in table["UPSTREAM_TABLES"]:
                upstream_table_fqdn = upstream_table["upstream_object_name"].split(".")
                if len(upstream_table_fqdn) != 3:
                    continue
                else:
                    upstream_global_id_arg = "{db}{schema}{table}".format(
                        db=upstream_table_fqdn[0], schema=upstream_table_fqdn[1], table=upstream_table_fqdn[2]
                    )
                    upstream_table_global_id = new_global_id(
                        tenant_id=tenant_id, cluster_id=endpoint, data_id=upstream_global_id_arg, data_type="table"
                    )
                    lineage_input.upstream.append(upstream_table_global_id)
            lineage_inputs = LineageInputs(
                downstream_global_id=downstream_table_global_id,
                downstream_database_name=downstream_table_fqdn[0],
                downstream_schema_name=downstream_table_fqdn[1],
                downstream_table_name=downstream_table_fqdn[2],
                downstream_column_name="",
                upstreams=lineage_input,
            )
            payload.append(lineage_inputs)
    return payload


def gen_column_lineage_payload(tenant_id: str, endpoint: str, columns: List[Dict[str, str]]) -> List[LineageInputs]:
    payload = list()
    for column in columns:
        downstream_table_fqdn = column["DOWNSTREAM_TABLE_NAME"].split(".")
        if len(downstream_table_fqdn) != 3:
            continue
        else:
            global_id_arg = "{db}{schema}{table}{column}".format(
                db=downstream_table_fqdn[0],
                schema=downstream_table_fqdn[1],
                table=downstream_table_fqdn[2],
                column=column["DOWNSTREAM_COLUMN_NAME"],
            )
            downstream_column_global_id = new_global_id(
                tenant_id=tenant_id, cluster_id=endpoint, data_id=global_id_arg, data_type="column"
            )
            upstream_columns: List[Dict[str, str]] = json.loads(column["UPSTREAM_COLUMNS"])
            lineage_input = LineageInput(upstream=[])
            for upstream_column in upstream_columns:
                upstream_table_fqdn = upstream_column["upstream_table_name"].split(".")
                if len(upstream_table_fqdn) != 3:
                    continue
                elif not upstream_column.get("upstream_column_name"):
                    continue
                else:
                    upstream_global_id_arg = "{db}{schema}{table}{column}".format(
                        db=upstream_table_fqdn[0],
                        schema=upstream_table_fqdn[1],
                        table=upstream_table_fqdn[2],
                        column=upstream_column["upstream_column_name"],
                    )
                    upstream_column_global_id = new_global_id(
                        tenant_id=tenant_id, cluster_id=endpoint, data_id=upstream_global_id_arg, data_type="column"
                    )
                    lineage_input.upstream.append(upstream_column_global_id)
            lineage_inputs = LineageInputs(
                downstream_global_id=downstream_column_global_id,
                downstream_database_name=downstream_table_fqdn[0],
                downstream_schema_name=downstream_table_fqdn[1],
                downstream_table_name=downstream_table_fqdn[2],
                downstream_column_name=column["DOWNSTREAM_COLUMN_NAME"],
                upstreams=lineage_input,
            )
            payload.append(lineage_inputs)
    return payload


def gen_table_lineage_payload_inputs(input_data: Tuple[List[str]]) -> List[Dict[str, Union[str, List[Dict[str, str]]]]]:
    result = {}

    for item in input_data:
        downstream_table_name = item[0]
        upstream_object_name = item[1]

        if downstream_table_name not in result:
            result[downstream_table_name] = {
                "DOWNSTREAM_TABLE_NAME": downstream_table_name,
                "UPSTREAM_TABLES": [],
            }

        result[downstream_table_name]["UPSTREAM_TABLES"].append(
            {
                "upstream_object_name": upstream_object_name,
            }
        )

    return list(result.values())


def parse_snowflake_results(results: List[Dict[str, str]]):
    payloads = list()
    for result in results:
        payload = dict()
        payload["DOWNSTREAM_TABLE_NAME"] = result["DOWNSTREAM_TABLE_NAME"]
        payload["DOWNSTREAM_TABLE_DOMAIN"] = result["DOWNSTREAM_TABLE_DOMAIN"]
        payload["UPSTREAM_TABLES"] = json.loads(result["UPSTREAM_TABLES"])
        payloads.append(payload)
    return payloads


def parse_databricks_table_lineage(results: List) -> List[Dict[str, Dict]]:
    # Parses results from Quollio Databricks lineage table
    # Returns tuple of downstream_table_name (0) and upstream_tables (1)
    payloads = list()
    for result in results:
        payload = dict()
        payload["DOWNSTREAM_TABLE_NAME"] = result["DOWNSTREAM_TABLE_NAME"]
        payload["UPSTREAM_TABLES"] = json.loads(result["UPSTREAM_TABLES"])
        payloads.append(payload)
    return payloads


def parse_bigquery_table_lineage(tables: Dict) -> List[Dict[str, Dict]]:
    payloads = list()
    for downstream, upstream in tables.items():
        payload = {
            "DOWNSTREAM_TABLE_NAME": "",
            "UPSTREAM_TABLES": [],
        }
        payload["DOWNSTREAM_TABLE_NAME"] = downstream
        for upstream_table in upstream:
            payload["UPSTREAM_TABLES"].append({"upstream_object_name": upstream_table})
        payloads.append(payload)
    return payloads



import inspect
import logging

LOG_LEVELS = {
    "critical": logging.CRITICAL,
    "error": logging.ERROR,
    "warning": logging.WARNING,
    "info": logging.INFO,
    "debug": logging.DEBUG,
    "notset": logging.NOTSET,
}

logger = logging.getLogger(__name__)


def configure_logging(level: str = "INFO"):
    """Configure logging settings."""
    log_level = LOG_LEVELS.get(level.lower())
    if log_level is None:
        raise ValueError(f"Unknown log level: {level}")

    logging.basicConfig(
        level=log_level,
        format="%(asctime)s - %(levelname)s - %(funcName)s:%(lineno)d - %(message)s",
    )
    logger.setLevel(log_level)
    logger.info(f"Logging is configured to {level} level.")


def error_handling_decorator(func):
    """Decorator for consistent error handling in CLI commands."""

    def wrapper(*args, **kwargs):
        func_name = func.__name__
        try:
            logger.debug(f"Starting {func_name}")
            result = func(*args, **kwargs)
            logger.debug(f"Completed {func_name} successfully")
            return result
        except Exception as e:
            current_frame = inspect.currentframe()
            error_frame = current_frame.f_back
            line_number = error_frame.f_lineno
            logger.error(f"Error in {func_name} at line {line_number}: {str(e)}", exc_info=True)
            raise

    return wrapper



from typing import Any, Dict, List

from google.cloud.bigquery import Client
from google.cloud.datacatalog_lineage_v1 import EntityReference, LineageClient, SearchLinksRequest
from google.oauth2.service_account import Credentials
from googleapiclient.discovery import build

from quollio_core.helper.log_utils import logger  # Importing the logger from logging_utils


class BigQueryClient:
    """Client to interact with the BigQuery API."""

    def __init__(self, credentials: Credentials, project_id: str) -> None:
        """Initialize the BigQuery client with provided credentials."""
        self.client = self.__initialize(credentials=credentials, project_id=project_id)

    def __initialize(self, credentials: Credentials, project_id: str) -> Client:
        return Client(credentials=credentials, project=project_id)

    def list_dataset_ids(self) -> List[str]:
        """List all dataset ids in the project."""
        datasets = list(self.client.list_datasets())
        logger.debug("Found %s datasets in project %s", len(datasets), self.client.project)
        return [dataset.dataset_id for dataset in datasets]

    def list_tables(self, dataset_id: str) -> List[Dict[str, str]]:
        """List all tables in the dataset."""
        tables = list(self.client.list_tables(dataset_id))
        logger.debug("Found %s tables in dataset %s", len(tables), dataset_id)
        return [
            {
                "table_id": table.table_id,
                "table_type": table.table_type,
                "project": table.project,
                "dataset_id": table.dataset_id,
            }
            for table in tables
        ]

    def get_columns(self, table_id: str, dataset_id: str) -> List[Dict[str, str]]:
        """Get the columns of the table."""
        table = self.client.get_table(f"{self.client.project}.{dataset_id}.{table_id}")
        return [{"name": field.name, "type": field.field_type} for field in table.schema]

    def get_all_columns(self) -> Dict[str, Dict[str, List[Dict[str, Any]]]]:
        """Get all columns in the project."""
        all_columns = {}
        datasets = self.list_dataset_ids()
        for dataset_id in datasets:
            all_columns[dataset_id] = {}
            tables = self.list_tables(dataset_id)
            for table_info in tables:
                table_id = table_info["table_id"]
                table_type = table_info["table_type"]
                columns = self.get_columns(table_id, dataset_id)
                all_columns[dataset_id][table_id] = {"columns": columns, "table_type": table_type}
        return all_columns


class GCPLineageClient:
    """Client to interact with the GCP Lineage API."""

    def __init__(self, credentials: Credentials) -> None:
        """Initialize the GCP Lineage client with provided credentials."""
        self.client = self.__initialze(credentials=credentials)

    def __initialze(self, credentials: Credentials) -> LineageClient:
        return LineageClient(credentials=credentials)

    def get_links(self, request: SearchLinksRequest) -> list:
        """Search for links between entities (tables)."""
        response = self.client.search_links(request)
        return response.links


def get_entitiy_reference() -> EntityReference:
    return EntityReference()


def get_search_request(downstream_table: EntityReference, project_id: str, region: str) -> SearchLinksRequest:
    return SearchLinksRequest(target=downstream_table, parent=f"projects/{project_id}/locations/{region.lower()}")


def get_credentials(credentials_json: dict) -> Credentials:
    return Credentials.from_service_account_info(credentials_json)


def get_org_id(credentials_json: dict) -> str:
    credentials = get_credentials(credentials_json)
    crm_service = build("cloudresourcemanager", "v1", credentials=credentials)
    project_id = credentials_json["project_id"]
    project = crm_service.projects().get(projectId=project_id).execute()
    return project["parent"]["id"]



from typing import Dict

import yaml
from blake3 import blake3
from jinja2 import Environment, FileSystemLoader


def new_global_id(tenant_id: str, cluster_id: str, data_id: str, data_type: str) -> str:
    prefix = ""
    data_types = {
        "schema": "schm-",
        "table": "tbl-",
        "column": "clmn-",
        "bigroup": "bgrp-",
        "dashboard": "dsbd-",
        "sheet": "sht-",
    }
    prefix = data_types[data_type]

    data_to_hash = f"{tenant_id}{cluster_id}{data_id}"
    hashed = blake3(data_to_hash.encode()).digest()
    global_id: str = prefix + hashed[:16].hex()
    return global_id


def setup_dbt_profile(connections_json: Dict[str, str], template_path: str, template_name: str) -> None:
    profile_path = f"{template_path}/profiles.yml"
    loader = Environment(loader=(FileSystemLoader(template_path, encoding="utf-8")))
    template = loader.get_template(template_name)
    profiles_body = template.render(connections_json)
    with open(profile_path, "w") as profiles:
        yaml.dump(yaml.safe_load(profiles_body), profiles, default_flow_style=False, allow_unicode=True)
    return


def trim_prefix(s: str, prefix: str) -> str:
    return s.lstrip(prefix)



import logging
from dataclasses import asdict, dataclass
from enum import Enum
from typing import Dict, List, Tuple

from redshift_connector import Connection, connect
from redshift_connector.error import ProgrammingError

logger = logging.getLogger(__name__)


# MEMO: https://docs.aws.amazon.com/redshift/latest/mgmt/rsql-query-tool-error-codes.html
class RedshiftErrorCode(Enum):
    # Connection
    CONNECTION_EXCEPTION = "08000"
    CONNECTION_DOES_NOT_EXIST = "08003"
    CONNECTION_FAILURE = "08006"
    SQLCLIENT_UNABLE_TO_ESTABLISH_SQLCONNECTION = "08001"
    SQLSERVER_REJECTED_ESTABLISHMENT_OF_SQLCONNECTION = "08004"
    TRANSACTION_RESOLUTION_UNKNOWN = "08007"
    PROTOCOL_VIOLATION = "08P01"

    # Privilege
    INSUFFICIENT_PRIVILEGE = "42501"

    # Resource
    OUT_OF_MEMORY = "53200"

    @classmethod
    def match_error_code(cls, error_code: str) -> bool:
        for code in cls:
            if code.value == error_code:
                return True
        return False


@dataclass
class RedshiftConnectionConfig:
    host: str
    build_user: str
    query_user: str
    build_password: str
    query_password: str
    database: str
    schema: str
    port: int = 5439
    threads: int = 3

    def as_dict(self) -> Dict[str, str]:
        return asdict(self)


class RedshiftQueryExecutor:
    def __init__(self, config: RedshiftConnectionConfig):
        self.conn = self.__initialize(config)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.conn.close()

    def __initialize(self, config: RedshiftConnectionConfig) -> Connection:
        conn: RedshiftConnectionConfig = connect(
            host=config.host, database=config.database, user=config.query_user, password=config.query_password
        )
        return conn

    def get_query_results(self, query: str) -> Tuple[List[str]]:
        with self.conn.cursor() as cur:
            try:
                cur.execute(query)
                result: tuple = cur.fetchall()
                return result
            except ProgrammingError as pe:
                err_body = dict()
                if 1 <= len(pe.args):
                    err_body = pe.args[0]

                error_code = err_body.get("C")
                if RedshiftErrorCode.match_error_code(error_code):
                    err = RedshiftErrorCode(error_code)
                    logger.error(" ".join(query.split()))
                    logger.error("{err} error happened: {body}".format(err=err.name, body=pe))
                else:
                    logger.error(query)
                    logger.error("ProgrammingError: {err}".format(err=pe))
                    raise
            except Exception as e:
                logger.error(query)
                logger.error("Failed to get query results. error: {err}".format(err=e))
                raise



"""
https://gist.github.com/orls/51525c86ee77a56ad396
Provides a utility to inject environment variables into argparse definitions.
Currently requires explicit naming of env vars to check for
"""

import argparse
import os
from distutils.util import strtobool
from typing import Union


# Courtesy of http://stackoverflow.com/a/10551190 with env-var retrieval fixed
class EnvDefault(argparse.Action):
    """An argparse action class that auto-sets missing default values from env
    vars. Defaults to requiring the argument."""

    def __init__(self, envvar, required=True, default=None, **kwargs):
        # override values if envvar exists
        if envvar in os.environ:
            if kwargs.get("nargs", None) is None:
                default = os.environ[envvar]
            else:
                default = os.environ[envvar].split(" ")
        if required and default:
            required = False
        super(EnvDefault, self).__init__(default=default, required=required, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)


class EnvStoreTrue(argparse._StoreTrueAction):
    """An argparse action class that auto-sets missing default values from env vars for store_true."""

    def __init__(self, envvar, required=True, default=None, **kwargs):
        # Only pass the arguments that argparse._StoreTrueAction expects
        action_kwargs = {key: value for key, value in kwargs.items() if key in ("option_strings", "dest")}
        if envvar in os.environ:
            default = _convert_value_to_bool(os.environ[envvar])
        if required and default:
            required = False
        super(EnvStoreTrue, self).__init__(default=default, required=required, **action_kwargs)


# functional sugar for the above
def env_default(envvar, store_true=False):
    def wrapper(**kwargs):
        if store_true:
            return EnvStoreTrue(envvar, **kwargs)
        return EnvDefault(envvar, **kwargs)

    return wrapper


def _convert_value_to_bool(v: Union[str, bool]) -> bool:
    if isinstance(v, str):
        return bool(strtobool(v))
    return v



from typing import Dict, List

from google.auth.credentials import Credentials

from quollio_core.helper.log_utils import error_handling_decorator, logger
from quollio_core.profilers.lineage import gen_table_lineage_payload, parse_bigquery_table_lineage
from quollio_core.profilers.stats import gen_table_stats_payload
from quollio_core.repository import qdc
from quollio_core.repository.bigquery import BigQueryClient, GCPLineageClient, get_entitiy_reference, get_search_request


@error_handling_decorator
def bigquery_table_lineage(
    qdc_client: qdc.QDCExternalAPIClient,
    tenant_id: str,
    project_id: str,
    regions: list,
    org_id: str,
    credentials: Credentials,
) -> None:
    lineage_client = GCPLineageClient(credentials)
    bq_client = BigQueryClient(credentials, project_id)

    datasets = bq_client.list_dataset_ids()
    all_tables = generate_table_list(bq_client, datasets)
    lineage_links = generate_lineage_links(all_tables, lineage_client, project_id, regions)
    lineage_links = parse_bigquery_table_lineage(lineage_links)
    logger.debug("The following resources will be ingested. %s", lineage_links)

    update_table_lineage_inputs = gen_table_lineage_payload(tenant_id=tenant_id, endpoint=org_id, tables=lineage_links)

    req_count = 0
    for update_table_lineage_input in update_table_lineage_inputs:
        logger.info(
            "Generating table lineage. downstream: %s -> %s -> %s. upstream: %s",
            update_table_lineage_input.downstream_database_name,
            update_table_lineage_input.downstream_schema_name,
            update_table_lineage_input.downstream_table_name,
            update_table_lineage_input.upstreams.as_dict(),
        )
        status_code = qdc_client.update_lineage_by_id(
            global_id=update_table_lineage_input.downstream_global_id,
            payload=update_table_lineage_input.upstreams.as_dict(),
        )
        if status_code == 200:
            req_count += 1
    logger.info("Generating table lineage is finished. %s lineages are ingested.", req_count)


@error_handling_decorator
def bigquery_table_stats(
    qdc_client: qdc.QDCExternalAPIClient,
    bq_client: BigQueryClient,
    tenant_id: str,
    org_id: str,
    dataplex_stats_tables: list,
) -> None:
    profiling_results = []
    for table in dataplex_stats_tables:
        logger.info("Profiling columns using Dataplex stats table: %s", table)
        profiling_results.extend(column_stats_from_dataplex(bq_client, table))

    stats = gen_table_stats_payload(tenant_id, org_id, profiling_results)

    for stat in stats:
        status_code = qdc_client.update_stats_by_id(
            global_id=stat.global_id,
            payload=stat.body.as_dict(),
        )
        if status_code == 200:
            logger.info(
                "Stats for column %s -> %s -> %s -> %s is successfully ingested.",
                stat.db,
                stat.schema,
                stat.table,
                stat.column,
            )
            logger.debug("Stats for column id %s is successfully ingested.", stat.global_id)


def generate_table_list(bq_client: BigQueryClient, datasets: List[str]) -> List[str]:
    all_tables = []
    for dataset in datasets:
        all_tables.extend(
            [
                table
                for table in bq_client.list_tables(dataset)
                if table["table_type"] in ["TABLE", "VIEW", "MATERIALIZED_VIEW"]
            ],
        )

    all_table_names = []
    for table in all_tables:
        all_table_names.append(f"{bq_client.client.project}.{table['dataset_id']}.{table['table_id']}")

    return all_table_names


def generate_lineage_links(
    all_tables: List[str],
    lineage_client: GCPLineageClient,
    project_id: str,
    regions: List[str],
) -> Dict[str, List[str]]:
    lineage_links = {}
    for table in all_tables:
        if "quollio" in table.lower():
            continue
        downstream = get_entitiy_reference()
        downstream.fully_qualified_name = f"bigquery:{table}"

        for region in regions:
            request = get_search_request(downstream_table=downstream, project_id=project_id, region=region)
            response = lineage_client.get_links(request=request)
            for lineage in response:
                target_table = str(lineage.target.fully_qualified_name).replace("bigquery:", "")
                source_table = str(lineage.source.fully_qualified_name).replace("bigquery:", "")
                if target_table not in lineage_links:
                    lineage_links[target_table] = []
                if source_table not in lineage_links[target_table]:
                    lineage_links[target_table].append(source_table)

    return lineage_links


def column_stats_from_dataplex(bq_client: BigQueryClient, profiling_table: str) -> List[Dict]:
    query = f"""
    SELECT
        data_source.table_project_id AS DB_NAME,
        data_source.dataset_id AS SCHEMA_NAME,
        data_source.table_id AS TABLE_NAME,
        column_name AS COLUMN_NAME,
        min_value AS MIN_VALUE,
        max_value AS MAX_VALUE,
        average_value AS AVG_VALUE,
        quartile_median AS MEDIAN_VALUE,
        standard_deviation AS STDDEV_VALUE,
        top_n[0][0] AS MODE_VALUE,
        CAST((percent_null / 100) * job_rows_scanned AS INT) as NULL_COUNT,
        CAST((percent_unique / 100) * job_rows_scanned AS INT) as CARDINALITY
    FROM `{profiling_table}`
    """
    logger.debug(f"Executing Query: {query}")
    results = bq_client.client.query(query).result()

    # Convert RowIterator to a list of dictionaries
    return [dict(row) for row in results]



import argparse
import logging
import os
import shutil

from quollio_core.helper.core import setup_dbt_profile
from quollio_core.helper.env_default import env_default
from quollio_core.helper.log import set_log_level
from quollio_core.profilers.snowflake import (
    snowflake_column_to_column_lineage,
    snowflake_table_level_sqllineage,
    snowflake_table_stats,
    snowflake_table_to_table_lineage,
)
from quollio_core.profilers.stats import get_column_stats_items
from quollio_core.repository import dbt, qdc, snowflake

logger = logging.getLogger(__name__)


def build_view(
    conn: snowflake.SnowflakeConnectionConfig,
    stats_sample_method: str,
    target_tables: str = "",
    log_level: str = "info",
    dbt_macro_source: str = "hub",
) -> None:
    logger.info("Build profiler views using dbt")
    # set parameters
    dbt_client = dbt.DBTClient()
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_path = f"{current_dir}/dbt_projects/snowflake"
    template_path = f"{current_dir}/dbt_projects/snowflake/profiles"
    template_name = "profiles_template.yml"
    options = '{{"query_role": {query_role}, "sample_method": {sample_method}}}'.format(
        query_role=conn.account_query_role,
        sample_method=stats_sample_method,
    )
    new_package_file = f"{project_path}/packages.yml"
    if dbt_macro_source == "local":
        shutil.copyfile(f"{project_path}/packages_local.yml", new_package_file)
        logger.info("Will install dbt macro defined in packages_local.yml")
    else:
        shutil.copyfile(f"{project_path}/packages_hub.yml", new_package_file)
        logger.info("Will install dbt macro defined in packages_hub.yml")

    # build views using dbt
    setup_dbt_profile(
        connections_json=conn.as_dict(),
        template_path=template_path,
        template_name=template_name,
    )
    # FIXME: when executing some of the commands, directory changes due to the library bug.
    # https://github.com/dbt-labs/dbt-core/issues/8997
    dbt_client.invoke(
        cmd="deps",
        project_dir=project_path,
        profile_dir=template_path,
        options=["--no-use-colors", "--log-level", log_level, "--vars", options, "--source", dbt_macro_source],
    )
    run_options = ["--no-use-colors", "--log-level", log_level, "--vars", options]
    if target_tables is not None:
        if "quollio_stats_columns" in target_tables:
            target_tables.append("quollio_stats_profiling_columns")
        target_tables_str = " ".join(target_tables)
        run_options.append("--select")
        run_options.append(target_tables_str)

    dbt_client.invoke(
        cmd="run",
        project_dir=project_path,
        profile_dir=template_path,
        options=run_options,
    )

    logger.info("Profiler views are successfully built.")
    return


def load_lineage(
    conn: snowflake.SnowflakeConnectionConfig,
    qdc_client: qdc.QDCExternalAPIClient,
    tenant_id: str,
    enable_column_lineage: bool = False,
) -> None:
    logger.info("Generate Snowflake table to table lineage.")

    snowflake_table_to_table_lineage(
        conn=conn,
        qdc_client=qdc_client,
        tenant_id=tenant_id,
    )

    if enable_column_lineage:
        logger.info(
            f"enable_column_lineage is set to {enable_column_lineage}.Generate Snowflake column to column lineage."
        )
        snowflake_column_to_column_lineage(
            conn=conn,
            qdc_client=qdc_client,
            tenant_id=tenant_id,
        )
    else:
        logger.info("Skip column lineage ingestion. Set enable_column_lineage to True if you ingest column lineage.")

    logger.info("Lineage data is successfully loaded.")

    return


def load_stats(
    conn: snowflake.SnowflakeConnectionConfig,
    qdc_client: qdc.QDCExternalAPIClient,
    tenant_id: str,
    stats_items: str,
) -> None:
    logger.info("Generate Snowflake stats.")

    if stats_items is None:
        raise ValueError("No stats items are not selected. Please specify any value to `stats_items` param.")

    logger.info("The following values will be aggregated. {stats_items}".format(stats_items=stats_items))
    snowflake_table_stats(
        conn=conn,
        qdc_client=qdc_client,
        tenant_id=tenant_id,
        stats_items=stats_items,
    )

    logger.info("Stats data is successfully loaded.")

    return


def load_sqllineage(
    conn: snowflake.SnowflakeConnectionConfig,
    qdc_client: qdc.QDCExternalAPIClient,
    tenant_id: str,
) -> None:
    logger.info("Generate Snowflake sqllineage.")
    snowflake_table_level_sqllineage(
        conn=conn,
        qdc_client=qdc_client,
        tenant_id=tenant_id,
    )

    logger.info("sqllineage data is successfully loaded.")

    return


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="Quollio Intelligence Agent for Snowflake",
        description="Build views and load lineage and stats to Quollio from Snowflake using dbt.",
        epilog="Copyright (c) 2024 Quollio Technologies, Inc.",
    )
    parser.add_argument(
        "commands",
        choices=["build_view", "load_lineage", "load_stats", "load_sqllineage"],
        type=str,
        nargs="+",
        help="""
        The command to execute.
        'build_view': Build views using dbt,
        'load_lineage': Load lineage data from created views to Quollio,
        'load_stats': Load stats from created views to Quollio,
        'load_sqllineage': Load lineage data from sql parse result(alpha),
        """,
    )
    parser.add_argument(
        "--account_id",
        type=str,
        action=env_default("SNOWFLAKE_ACCOUNT_ID"),
        required=False,
        help="The target account ID of Snowflake",
    )
    parser.add_argument(
        "--user",
        type=str,
        action=env_default("SNOWFLAKE_USER"),
        required=False,
        help="User name to connect to the database",
    )
    parser.add_argument(
        "--password",
        type=str,
        action=env_default("SNOWFLAKE_PASSWORD"),
        required=False,
        help="User password to connect to the database",
    )
    parser.add_argument(
        "--build_role",
        type=str,
        action=env_default("SNOWFLAKE_BUILD_ROLE"),
        required=False,
        help="The role in Snowflake that is used to build views by dbt",
    )
    parser.add_argument(
        "--query_role",
        type=str,
        action=env_default("SNOWFLAKE_QUERY_ROLE"),
        required=False,
        help="The role in Snowflake that is used to query views",
    )
    parser.add_argument(
        "--warehouse",
        type=str,
        action=env_default("SNOWFLAKE_WAREHOUSE"),
        required=False,
        help="The warehouse in Snowflake that is used to execute statements",
    )
    parser.add_argument(
        "--database",
        type=str,
        action=env_default("SNOWFLAKE_TARGET_DATABASE"),
        required=False,
        help="Target database name where the views are built by dbt",
    )
    parser.add_argument(
        "--schema",
        type=str,
        action=env_default("SNOWFLAKE_TARGET_SCHEMA"),
        default="public",
        required=False,
        help="Target schema name where the views are built by dbt",
    )
    parser.add_argument(
        "--target_tables",
        type=str,
        nargs="*",
        choices=["quollio_lineage_column_level", "quollio_lineage_table_level", "quollio_stats_columns"],
        action=env_default("SNOWFLAKE_TARGET_TABLES"),
        required=False,
        help="Target table name if you want to create only specific tables. \
              You need to specify this parameter if you want to specify tables, not all ones. \
              Please specify table name with blank delimiter like tableA tableB \
              if you want to create two or more tables.",
    )
    parser.add_argument(
        "--sample_method",
        type=str,
        action=env_default("SNOWFLAKE_SAMPLE_METHOD"),
        default="SAMPLE(0.0001)",
        required=False,
        help="The method to sample data for stats",
    )
    parser.add_argument(
        "--tenant_id",
        type=str,
        action=env_default("TENANT_ID"),
        required=False,
        help="The tenant id (company id) where the lineage and stats are loaded",
    )
    parser.add_argument(
        "--log_level",
        type=str,
        choices=["debug", "info", "warn", "error", "none"],
        action=env_default("LOG_LEVEL"),
        required=False,
        help="The log level for dbt commands. Default value is info",
    )
    parser.add_argument(
        "--dbt_macro_source",
        type=str,
        choices=["hub", "local"],
        action=env_default("DBT_MACRO_SOURCE"),
        default="hub",
        required=False,
        help="The dbt macro source",
    )
    parser.add_argument(
        "--api_url",
        type=str,
        action=env_default("QDC_API_URL"),
        required=False,
        help="The base URL of Quollio External API",
    )
    parser.add_argument(
        "--client_id",
        type=str,
        action=env_default("QDC_CLIENT_ID"),
        required=False,
        help="The client id that is created on Quollio console to let clients access Quollio External API",
    )
    parser.add_argument(
        "--client_secret",
        type=str,
        action=env_default("QDC_CLIENT_SECRET"),
        required=False,
        help="The client secrete that is created on Quollio console to let clients access Quollio External API",
    )
    parser.add_argument(
        "--enable_column_lineage",
        type=bool,
        action=env_default("ENABLE_COLUMN_LINEAGE", store_true=True),
        default=False,
        required=False,
        help="Whether to ingest column lineage into QDIC or not. Default value is False",
    )

    stats_items = get_column_stats_items()
    parser.add_argument(
        "--target_stats_items",
        type=str,
        nargs="*",
        choices=stats_items,
        default=stats_items,
        action=env_default("SNOWFLAKE_STATS_ITEMS"),
        required=False,
        help="The items for statistic values.\
              You can choose the items to be aggregated for stats. All items are selected by default.",
    )
    args = parser.parse_args()
    set_log_level(level=args.log_level)

    conn = snowflake.SnowflakeConnectionConfig(
        account_id=args.account_id,
        account_user=args.user,
        account_password=args.password,
        account_build_role=args.build_role,
        account_query_role=args.query_role,
        account_warehouse=args.warehouse,
        account_database=args.database,
        account_schema=args.schema,
    )

    if len(args.commands) == 0:
        raise ValueError("No command is provided")

    if "build_view" in args.commands:
        build_view(
            conn=conn,
            stats_sample_method=args.sample_method,
            target_tables=args.target_tables,
            log_level=args.log_level,
            dbt_macro_source=args.dbt_macro_source,
        )
    if "load_lineage" in args.commands:
        qdc_client = qdc.QDCExternalAPIClient(
            base_url=args.api_url,
            client_id=args.client_id,
            client_secret=args.client_secret,
        )
        load_lineage(
            conn=conn,
            qdc_client=qdc_client,
            tenant_id=args.tenant_id,
            enable_column_lineage=args.enable_column_lineage,
        )
    if "load_stats" in args.commands:
        qdc_client = qdc.QDCExternalAPIClient(
            base_url=args.api_url,
            client_id=args.client_id,
            client_secret=args.client_secret,
        )
        load_stats(
            conn=conn,
            qdc_client=qdc_client,
            tenant_id=args.tenant_id,
            stats_items=args.target_stats_items,
        )
    if "load_sqllineage" in args.commands:
        qdc_client = qdc.QDCExternalAPIClient(
            base_url=args.api_url,
            client_id=args.client_id,
            client_secret=args.client_secret,
        )
        load_sqllineage(
            conn=conn,
            qdc_client=qdc_client,
            tenant_id=args.tenant_id,
        )



import logging
from typing import Dict, List

from quollio_core.profilers.lineage import (
    gen_column_lineage_payload,
    gen_table_lineage_payload,
    parse_databricks_table_lineage,
)
from quollio_core.profilers.stats import gen_table_stats_payload, get_is_target_stats_items, render_sql_for_stats
from quollio_core.repository import databricks, qdc

logger = logging.getLogger(__name__)


def databricks_table_level_lineage(
    conn: databricks.DatabricksConnectionConfig,
    endpoint: str,
    qdc_client: qdc.QDCExternalAPIClient,
    tenant_id: str,
    dbt_table_name: str = "quollio_lineage_table_level",
) -> None:
    with databricks.DatabricksQueryExecutor(config=conn) as databricks_executor:
        results = databricks_executor.get_query_results(
            query=f"""
            SELECT
                DOWNSTREAM_TABLE_NAME,
                UPSTREAM_TABLES
            FROM {conn.catalog}.{conn.schema}.{dbt_table_name}
            """
        )
        tables = parse_databricks_table_lineage(results)
        update_table_lineage_inputs = gen_table_lineage_payload(
            tenant_id=tenant_id,
            endpoint=endpoint,
            tables=tables,
        )

        req_count = 0
        for update_table_lineage_input in update_table_lineage_inputs:
            logger.info(
                "Generating table lineage. downstream: %s -> %s-> %s",
                update_table_lineage_input.downstream_database_name,
                update_table_lineage_input.downstream_schema_name,
                update_table_lineage_input.downstream_table_name,
            )
            status_code = qdc_client.update_lineage_by_id(
                global_id=update_table_lineage_input.downstream_global_id,
                payload=update_table_lineage_input.upstreams.as_dict(),
            )
            if status_code == 200:
                req_count += 1
        logger.info("Generating table lineage is finished. %s lineages are ingested.", req_count)
    return


def databricks_column_level_lineage(
    conn: databricks.DatabricksConnectionConfig,
    endpoint: str,
    qdc_client: qdc.QDCExternalAPIClient,
    tenant_id: str,
    dbt_table_name: str = "quollio_lineage_column_level",
) -> None:
    with databricks.DatabricksQueryExecutor(config=conn) as databricks_executor:
        results = databricks_executor.get_query_results(
            query=f"""
            SELECT
                *
            FROM
                {conn.catalog}.{conn.schema}.{dbt_table_name}
            """
        )

    update_column_lineage_inputs = gen_column_lineage_payload(
        tenant_id=tenant_id,
        endpoint=endpoint,
        columns=results,
    )

    req_count = 0
    for update_column_lineage_input in update_column_lineage_inputs:
        logger.info(
            "Generating column lineage. downstream: %s -> %s -> %s -> %s",
            update_column_lineage_input.downstream_database_name,
            update_column_lineage_input.downstream_schema_name,
            update_column_lineage_input.downstream_table_name,
            update_column_lineage_input.downstream_column_name,
        )
        status_code = qdc_client.update_lineage_by_id(
            global_id=update_column_lineage_input.downstream_global_id,
            payload=update_column_lineage_input.upstreams.as_dict(),
        )
        if status_code == 200:
            req_count += 1
    logger.info(
        "Generating column lineage is finished. %s lineages are ingested.",
        req_count,
    )
    return


def _get_monitoring_tables(
    conn: databricks.DatabricksConnectionConfig, monitoring_table_suffix: str = "_profile_metrics"
) -> List[Dict[str, str]]:
    tables = []
    query = f"""
        SELECT
            table_catalog,
            table_schema,
            table_name,
            CONCAT(table_catalog, '.', table_schema, '.', table_name) AS table_fqdn
        FROM
            system.information_schema.tables
        WHERE
            table_name LIKE "%{monitoring_table_suffix}"
            AND table_name NOT LIKE ('quollio_%')
        """
    with databricks.DatabricksQueryExecutor(config=conn) as databricks_executor:
        tables = databricks_executor.get_query_results(query)
    if len(tables) > 0:
        logger.info("Found %s monitoring tables.", len(tables))
        return tables
    else:
        logger.info("No monitoring tables found.")
        return []


def _get_column_stats(
    conn: databricks.DatabricksConnectionConfig,
    stats_items: List[str],
    monitoring_table_suffix: str = "_profile_metrics",
) -> List[Dict[str, str]]:
    tables = _get_monitoring_tables(conn, monitoring_table_suffix)
    if not tables:
        return []
    stats = []
    is_aggregate_items = get_is_target_stats_items(stats_items=stats_items)
    for table in tables:
        monitored_table = table["table_fqdn"].removesuffix("_profile_metrics")
        monitored_table = monitored_table.split(".")
        if len(monitored_table) != 3:
            raise ValueError(f"Invalid table name: {table['table_fqdn']}")
        with databricks.DatabricksQueryExecutor(config=conn) as databricks_executor:
            cte = """
            WITH profile_record_history AS (
                SELECT
                    COLUMN_NAME
                    , distinct_count as cardinality
                    , MAX as max_value
                    , MIN as min_value
                    , AVG as avg_value
                    , MEDIAN as median_value
                    , STDDEV as stddev_value
                    , NUM_NULLS as null_count
                    , get(frequent_items, 0).item AS mode_value
                    , row_number() over(partition by column_name order by window desc) rownum
                FROM
                    {monitoring_table}
                WHERE
                    column_name not in (':table')
            ), profile_record AS (
            SELECT
                "{monitored_table_catalog}" as db_name
                , "{monitored_table_schema}" as schema_name
                , "{monitored_table_name}" as table_name
                , column_name
                , max_value
                , min_value
                , null_count
                , cardinality
                , avg_value
                , median_value
                , mode_value
                , stddev_value
            FROM
                profile_record_history
            WHERE
                rownum = 1
            )""".format(
                monitoring_table=table["table_fqdn"],
                monitored_table_catalog=monitored_table[0],
                monitored_table_schema=monitored_table[1],
                monitored_table_name=monitored_table[2],
            )
            query = render_sql_for_stats(is_aggregate_items=is_aggregate_items, table_fqn="profile_record", cte=cte)
            logger.debug(f"The following sql will be fetched to retrieve stats values. {query}")
            stats.append(databricks_executor.get_query_results(query))
    return stats


def databricks_column_stats(
    conn: databricks.DatabricksConnectionConfig,
    endpoint: str,
    qdc_client: qdc.QDCExternalAPIClient,
    tenant_id: str,
    stats_items: List[str],
    monitoring_table_suffix: str = "_profile_metrics",
) -> None:
    table_stats = _get_column_stats(conn, stats_items, monitoring_table_suffix)
    for table in table_stats:
        logger.debug("Table %s will be aggregated.", table)
        stats = gen_table_stats_payload(tenant_id=tenant_id, endpoint=endpoint, stats=table)
        for stat in stats:
            status_code = qdc_client.update_stats_by_id(
                global_id=stat.global_id,
                payload=stat.body.as_dict(),
            )
            if status_code == 200:
                logger.info("Stats for %s is successfully ingested.", stat.global_id)
    return



import logging
from typing import List

from quollio_core.profilers.lineage import (
    gen_column_lineage_payload,
    gen_table_lineage_payload,
    parse_snowflake_results,
)
from quollio_core.profilers.sqllineage import SQLLineage
from quollio_core.profilers.stats import gen_table_stats_payload, get_is_target_stats_items, render_sql_for_stats
from quollio_core.repository import qdc, snowflake

logger = logging.getLogger(__name__)


def snowflake_table_to_table_lineage(
    conn: snowflake.SnowflakeConnectionConfig,
    qdc_client: qdc.QDCExternalAPIClient,
    tenant_id: str,
) -> None:
    with snowflake.SnowflakeQueryExecutor(conn) as sf_executor:
        results = sf_executor.get_query_results(
            query="""
            SELECT
                *
            FROM
               {db}.{schema}.QUOLLIO_LINEAGE_TABLE_LEVEL
            """.format(
                db=conn.account_database,
                schema=conn.account_schema,
            )
        )
        parsed_results = parse_snowflake_results(results=results)
        update_table_lineage_inputs = gen_table_lineage_payload(
            tenant_id=tenant_id,
            endpoint=conn.account_id,
            tables=parsed_results,
        )

        req_count = 0
        for update_table_lineage_input in update_table_lineage_inputs:
            logger.info(
                "Generating table lineage. downstream: {db} -> {schema} -> {table}".format(
                    db=update_table_lineage_input.downstream_database_name,
                    schema=update_table_lineage_input.downstream_schema_name,
                    table=update_table_lineage_input.downstream_table_name,
                )
            )
            status_code = qdc_client.update_lineage_by_id(
                global_id=update_table_lineage_input.downstream_global_id,
                payload=update_table_lineage_input.upstreams.as_dict(),
            )
            if status_code == 200:
                req_count += 1
        logger.info(f"Generating table lineage is finished. {req_count} lineages are ingested.")
    return


def snowflake_column_to_column_lineage(
    conn: snowflake.SnowflakeConnectionConfig,
    qdc_client: qdc.QDCExternalAPIClient,
    tenant_id: str,
) -> None:
    with snowflake.SnowflakeQueryExecutor(conn) as sf_executor:
        results = sf_executor.get_query_results(
            query="""
            SELECT
                *
            FROM
                {db}.{schema}.QUOLLIO_LINEAGE_COLUMN_LEVEL
            """.format(
                db=conn.account_database,
                schema=conn.account_schema,
            )
        )
        update_column_lineage_inputs = gen_column_lineage_payload(
            tenant_id=tenant_id,
            endpoint=conn.account_id,
            columns=results,
        )

        req_count = 0
        for update_column_lineage_input in update_column_lineage_inputs:
            logger.info(
                "Generating column lineage. downstream: {db} -> {schema} -> {table} -> {column}".format(
                    db=update_column_lineage_input.downstream_database_name,
                    schema=update_column_lineage_input.downstream_schema_name,
                    table=update_column_lineage_input.downstream_table_name,
                    column=update_column_lineage_input.downstream_column_name,
                )
            )
            status_code = qdc_client.update_lineage_by_id(
                global_id=update_column_lineage_input.downstream_global_id,
                payload=update_column_lineage_input.upstreams.as_dict(),
            )
            if status_code == 200:
                req_count += 1
        logger.info(f"Generating column lineage is finished. {req_count} lineages are ingested.")
    return


def snowflake_table_level_sqllineage(
    conn: snowflake.SnowflakeConnectionConfig,
    qdc_client: qdc.QDCExternalAPIClient,
    tenant_id: str,
) -> None:
    with snowflake.SnowflakeQueryExecutor(conn) as sf_executor:
        results = sf_executor.get_query_results(
            query="""
            SELECT
                database_name
                , schema_name
                , query_text
            FROM
               {db}.{schema}.QUOLLIO_SQLLINEAGE_SOURCES
            """.format(
                db=conn.account_database,
                schema=conn.account_schema,
            )
        )
        update_table_lineage_inputs_list = list()
        sql_lineage = SQLLineage()
        for result in results:
            src_tables, dest_table = sql_lineage.get_table_level_lineage_source(
                sql=result["QUERY_TEXT"],
                dialect="snowflake",
                dest_db=result["DATABASE_NAME"],
                dest_schema=result["SCHEMA_NAME"],
            )
            update_table_lineage_inputs = sql_lineage.gen_lineage_input(
                tenant_id=tenant_id, endpoint=conn.account_id, src_tables=src_tables, dest_table=dest_table
            )
            update_table_lineage_inputs_list.append(update_table_lineage_inputs)

        req_count = 0
        for update_table_lineage_input in update_table_lineage_inputs_list:
            logger.info(
                "Generating table lineage. downstream: {db} -> {schema} -> {table}".format(
                    db=update_table_lineage_input.downstream_database_name,
                    schema=update_table_lineage_input.downstream_schema_name,
                    table=update_table_lineage_input.downstream_table_name,
                )
            )
            status_code = qdc_client.update_lineage_by_id(
                global_id=update_table_lineage_input.downstream_global_id,
                payload=update_table_lineage_input.upstreams.as_dict(),
            )
            if status_code == 200:
                req_count += 1
        logger.info(f"Generating table lineage is finished. {req_count} lineages are ingested.")
    return


def snowflake_table_stats(
    conn: snowflake.SnowflakeConnectionConfig,
    qdc_client: qdc.QDCExternalAPIClient,
    tenant_id: str,
    stats_items: List[str],
) -> None:
    with snowflake.SnowflakeQueryExecutor(conn) as sf_executor:
        stats_query = _gen_get_stats_views_query(
            db=conn.account_database,
            schema=conn.account_schema,
        )
        stats_views = sf_executor.get_query_results(query=stats_query)

        req_count = 0
        is_aggregate_items = get_is_target_stats_items(stats_items=stats_items)
        for stats_view in stats_views:
            table_fqn = "{catalog}.{schema}.{table}".format(
                catalog=stats_view["TABLE_CATALOG"], schema=stats_view["TABLE_SCHEMA"], table=stats_view["TABLE_NAME"]
            )
            stats_query = render_sql_for_stats(is_aggregate_items=is_aggregate_items, table_fqn=table_fqn)
            logger.debug(f"The following sql will be fetched to retrieve stats values. {stats_query}")
            stats_result = sf_executor.get_query_results(query=stats_query)
            payloads = gen_table_stats_payload(tenant_id=tenant_id, endpoint=conn.account_id, stats=stats_result)
            for payload in payloads:
                logger.info(
                    "Generating table stats. asset: {db} -> {schema} -> {table} -> {column}".format(
                        db=payload.db,
                        schema=payload.schema,
                        table=payload.table,
                        column=payload.column,
                    )
                )
                status_code = qdc_client.update_stats_by_id(
                    global_id=payload.global_id,
                    payload=payload.body.get_column_stats(),
                )
                if status_code == 200:
                    req_count += 1
        logger.info(f"Generating table stats is finished. {req_count} stats are ingested.")
    return


def _gen_get_stats_views_query(db: str, schema: str) -> str:
    query = """
        SELECT
            DISTINCT
            TABLE_CATALOG
            , TABLE_SCHEMA
            , TABLE_NAME
        FROM
            {db}.INFORMATION_SCHEMA.TABLES
        WHERE
            startswith(TABLE_NAME, 'QUOLLIO_STATS_COLUMNS_')
            AND TABLE_SCHEMA = UPPER('{schema}')
        """.format(
        db=db, schema=schema
    )
    return query



import logging
from dataclasses import asdict, dataclass, fields
from decimal import ROUND_HALF_UP, Decimal
from typing import Dict, List, Tuple, Union

from jinja2 import Template

from quollio_core.helper.core import new_global_id

logger = logging.getLogger(__name__)


@dataclass
class ColumnStatsInput:
    cardinality: int
    max: str
    mean: str
    median: str
    min: str
    mode: str
    number_of_null: int
    number_of_unique: int
    stddev: str

    def as_dict(self) -> Dict[str, str]:
        return asdict(self)


@dataclass
class TableStatsInput:
    count: int
    size: float

    def as_dict(self) -> Dict[str, str]:
        return asdict(self)


@dataclass
class StatsInput:
    column_stats: ColumnStatsInput
    table_stats: TableStatsInput

    def as_dict(self) -> Dict[str, str]:
        return asdict(self)

    def get_column_stats(self):
        return {"column_stats": self.column_stats.as_dict()}


@dataclass
class StatsRequest:
    global_id: str
    db: str
    schema: str
    table: str
    column: str
    body: StatsInput

    def as_dict(self) -> Dict[str, str]:
        return asdict(self)


def convert_value_type(obj, cast_str: bool = False):
    if obj is None:
        return None
    if isinstance(obj, Decimal):
        return str(obj.quantize(Decimal("0.1"), ROUND_HALF_UP))
    if cast_str:
        return str(obj)
    return obj


def gen_table_stats_payload(tenant_id: str, endpoint: str, stats: List[Dict[str, str]]) -> List[StatsRequest]:
    payloads = list()
    for stat in stats:
        db_name = stat.get("DB_NAME", stat.get("db_name"))
        schema_name = stat.get("SCHEMA_NAME", stat.get("schema_name"))
        table_name = stat.get("TABLE_NAME", stat.get("table_name"))
        column_name = stat.get("COLUMN_NAME", stat.get("column_name"))
        global_id_arg = "{db}{schema}{table}{column}".format(
            db=db_name, schema=schema_name, table=table_name, column=column_name
        )
        table_global_id = new_global_id(
            tenant_id=tenant_id, cluster_id=endpoint, data_id=global_id_arg, data_type="column"
        )
        column_stats_input = ColumnStatsInput(
            cardinality=convert_value_type(stat.get("CARDINALITY", stat.get("cardinality"))),
            max=convert_value_type(stat.get("MAX_VALUE", stat.get("max_value")), True),
            mean=convert_value_type(stat.get("AVG_VALUE", stat.get("avg_value")), True),
            median=convert_value_type(stat.get("MEDIAN_VALUE", stat.get("median_value")), True),
            min=convert_value_type(stat.get("MIN_VALUE", stat.get("min_value")), True),
            mode=convert_value_type(stat.get("MODE_VALUE", stat.get("mode_value")), True),
            number_of_null=convert_value_type(stat.get("NULL_COUNT", stat.get("null_count"))),
            number_of_unique=convert_value_type(stat.get("CARDINALITY", stat.get("cardinality"))),
            stddev=convert_value_type(stat.get("STDDEV_VALUE", stat.get("stddev_value")), True),
        )
        stats_request = StatsRequest(
            global_id=table_global_id,
            db=db_name,
            schema=schema_name,
            table=table_name,
            column=column_name,
            body=StatsInput(
                column_stats=column_stats_input,
                # MEMO: Table stats can be collected with metadata agent.
                # Then, It's not necessary to update with this system for now.
                table_stats=TableStatsInput(count=0, size=0.0),
            ),
        )
        payloads.append(stats_request)
    return payloads


def gen_table_stats_payload_from_tuple(
    tenant_id: str, endpoint: str, stats: Tuple[List[str]]
) -> List[Dict[str, Union[str, List[Dict[str, str]]]]]:
    payloads = list()
    for stat in stats:
        global_id_arg = "{db}{schema}{table}{column}".format(db=stat[0], schema=stat[1], table=stat[2], column=stat[3])
        table_global_id = new_global_id(
            tenant_id=tenant_id, cluster_id=endpoint, data_id=global_id_arg, data_type="column"
        )
        stats_request = StatsRequest(
            global_id=table_global_id,
            db=stat[0],
            schema=stat[1],
            table=stat[2],
            column=stat[3],
            body=StatsInput(
                column_stats=ColumnStatsInput(
                    max=convert_value_type(stat[4], True),
                    min=convert_value_type(stat[5], True),
                    number_of_null=convert_value_type(stat[6]),
                    cardinality=convert_value_type(stat[7]),
                    mean=convert_value_type(stat[8], True),
                    median=convert_value_type(stat[9], True),
                    mode=convert_value_type(stat[10], True),
                    number_of_unique=convert_value_type(stat[7]),
                    stddev=convert_value_type(stat[11], True),
                ),
                # MEMO: Table stats can be collected with metadata agent.
                # Then, It's not necessary to update with this system for now.
                table_stats=TableStatsInput(count=0, size=0.0),
            ),
        )
        payloads.append(stats_request)
    return payloads


def render_sql_for_stats(is_aggregate_items: Dict[str, bool], table_fqn: str, cte: str = "") -> str:
    sql_template_for_stats = Template(
        """
        {% if cte -%}
          {{ cte }}
        {% endif -%}
        SELECT
            db_name
            , schema_name
            , table_name
            , column_name
            , {% if agg_max == True -%} max_value {% else -%} null as max_value {% endif %}
            , {% if agg_min == True -%} min_value {% else -%} null as min_value {% endif %}
            , {% if agg_null_count == True -%} null_count {% else -%} null as null_count {% endif %}
            , {% if agg_cardinality == True -%} cardinality {% else -%} null as cardinality {% endif %}
            , {% if agg_avg == True -%} avg_value {% else -%} null as avg_value {% endif %}
            , {% if agg_median == True -%} median_value {% else -%} null as median_value {% endif %}
            , {% if agg_mode == True -%} mode_value {% else -%} null as mode_value {% endif %}
            , {% if agg_stddev == True -%} stddev_value {% else -%} null as stddev_value {% endif %}
        FROM
            {{ table_fqn }}
    """
    )
    query = sql_template_for_stats.render(
        agg_max=is_aggregate_items["max"],
        agg_min=is_aggregate_items["min"],
        agg_null_count=is_aggregate_items["number_of_null"],
        agg_cardinality=is_aggregate_items["cardinality"],
        agg_avg=is_aggregate_items["mean"],
        agg_median=is_aggregate_items["median"],
        agg_mode=is_aggregate_items["mode"],
        agg_stddev=is_aggregate_items["stddev"],
        table_fqn=table_fqn,
        cte=cte,
    )
    return query


def get_is_target_stats_items(stats_items: List[str]) -> List[Dict[str, bool]]:
    target_stats_fields = get_column_stats_items()
    is_aggregate_items = dict()
    for target_stats_field in target_stats_fields:
        is_aggregate_items[target_stats_field] = False

    for stats_item in stats_items:
        is_aggregate_items[stats_item] = True

    return is_aggregate_items


def get_column_stats_items() -> List[str]:
    return [field.name for field in fields(ColumnStatsInput)]



import logging


def set_log_level(level: str = "info") -> None:
    fmt = "%(asctime)s - %(levelname)s - %(name)s - %(message)s"
    if level == "info":
        logging.basicConfig(level=logging.INFO, format=fmt)
    elif level == "debug":
        logging.basicConfig(level=logging.DEBUG, format=fmt)
    elif level == "warn":
        logging.basicConfig(level=logging.WARNING, format=fmt)
    elif level == "error":
        logging.basicConfig(level=logging.ERROR, format=fmt)
    elif level == "critical":
        logging.basicConfig(level=logging.CRITICAL, format=fmt)
    else:
        logging.basicConfig(level=logging.NOTSET, format=fmt)
